@@ -14,10 +14,8 @@
 
 #include "ClangdLSPServer.h"
 #include "ClangdServer.h"
-#include "CodeComplete.h"
-#include "FSProvider.h"
+#include "support/ThreadsafeFS.h"
 #include <cstdio>
-#include <sstream>
 
 using namespace clang::clangd;
 
@@ -30,14 +28,14 @@ extern "C" int LLVMFuzzerTestOneInput(uint8_t *data, size_t size) {
   auto Transport = newJSONTransport(In, llvm::nulls(),
                                     /*InMirror=*/nullptr, /*Pretty=*/false,
                                     /*Style=*/JSONStreamStyle::Delimited);
-  RealFileSystemProvider FS;
+  RealThreadsafeFS FS;
   CodeCompleteOptions CCOpts;
-  CCOpts.EnableSnippets = false;
-  ClangdServer::Options Opts;
+  ClangdLSPServer::Options Opts;
+  Opts.CodeComplete.EnableSnippets = false;
+  Opts.UseDirBasedCDB = false;
 
   // Initialize and run ClangdLSPServer.
-  ClangdLSPServer LSPServer(*Transport, FS, CCOpts, llvm::None, false,
-                            llvm::None, Opts);
+  ClangdLSPServer LSPServer(*Transport, FS, Opts);
   LSPServer.run();
   return 0;
 }
