@@ -8,25 +8,27 @@
 
 #include "index/Background.h"
 #include "Compiler.h"
-#include "Context.h"
-#include "FSProvider.h"
+#include "Config.h"
 #include "Headers.h"
-#include "Logger.h"
 #include "ParsedAST.h"
-#include "Path.h"
 #include "SourceCode.h"
 #include "Symbol.h"
-#include "Threading.h"
-#include "Trace.h"
 #include "URI.h"
 #include "index/BackgroundIndexLoader.h"
 #include "index/FileIndex.h"
+#include "index/Index.h"
 #include "index/IndexAction.h"
 #include "index/MemIndex.h"
 #include "index/Ref.h"
 #include "index/Relation.h"
 #include "index/Serialization.h"
 #include "index/SymbolCollector.h"
+#include "support/Context.h"
+#include "support/Logger.h"
+#include "support/Path.h"
+#include "support/Threading.h"
+#include "support/ThreadsafeFS.h"
+#include "support/Trace.h"
 #include "clang/Basic/SourceLocation.h"
 #include "clang/Basic/SourceManager.h"
 #include "clang/Driver/Types.h"
@@ -41,6 +43,7 @@
 #include "llvm/Support/Error.h"
 #include "llvm/Support/Path.h"
 #include "llvm/Support/Threading.h"
+#include "llvm/Support/xxhash.h"
 
 #include <algorithm>
 #include <atomic>
@@ -61,51 +64,6 @@ namespace clang {
 namespace clangd {
 namespace {
 
-// Resolves URI to file paths with cache.
-class URIToFileCache {
-public:
-  URIToFileCache(llvm::StringRef HintPath) : HintPath(HintPath) {}
-
-  llvm::StringRef resolve(llvm::StringRef FileURI) {
-    auto I = URIToPathCache.try_emplace(FileURI);
-    if (I.second) {
-      auto Path = URI::resolve(FileURI, HintPath);
-      if (!Path) {
-        elog("Failed to resolve URI {0}: {1}", FileURI, Path.takeError());
-        assert(false && "Failed to resolve URI");
-        return "";
-      }
-      I.first->second = *Path;
-    }
-    return I.first->second;
-  }
-
-private:
-  std::string HintPath;
-  llvm::StringMap<std::string> URIToPathCache;
-};
-
-// We keep only the node "U" and its edges. Any node other than "U" will be
-// empty in the resultant graph.
-IncludeGraph getSubGraph(const URI &U, const IncludeGraph &FullGraph) {
-  IncludeGraph IG;
-
-  std::string FileURI = U.toString();
-  auto Entry = IG.try_emplace(FileURI).first;
-  auto &Node = Entry->getValue();
-  Node = FullGraph.lookup(Entry->getKey());
-  Node.URI = Entry->getKey();
-
-  // URIs inside nodes must point into the keys of the same IncludeGraph.
-  for (auto &Include : Node.DirectIncludes) {
-    auto I = IG.try_emplace(Include).first;
-    I->getValue().URI = I->getKey();
-    Include = I->getKey();
-  }
-
-  return IG;
-}
-
 // We cannot use vfs->makeAbsolute because Cmd.FileName is either absolute or
 // relative to Cmd.Directory, which might not be the same as current working
 // directory.
@@ -135,24 +93,26 @@ bool shardIsStale(const LoadedShard &LS, llvm::vfs::FileSystem *FS) {
 } // namespace
 
 BackgroundIndex::BackgroundIndex(
-    Context BackgroundContext, const FileSystemProvider &FSProvider,
-    const GlobalCompilationDatabase &CDB,
-    BackgroundIndexStorage::Factory IndexStorageFactory, size_t ThreadPoolSize)
-    : SwapIndex(std::make_unique<MemIndex>()), FSProvider(FSProvider),
-      CDB(CDB), BackgroundContext(std::move(BackgroundContext)),
-      Rebuilder(this, &IndexedSymbols, ThreadPoolSize),
+    const ThreadsafeFS &TFS, const GlobalCompilationDatabase &CDB,
+    BackgroundIndexStorage::Factory IndexStorageFactory, Options Opts)
+    : SwapIndex(std::make_unique<MemIndex>()), TFS(TFS), CDB(CDB),
+      ContextProvider(std::move(Opts.ContextProvider)),
+      IndexedSymbols(IndexContents::All),
+      Rebuilder(this, &IndexedSymbols, Opts.ThreadPoolSize),
       IndexStorageFactory(std::move(IndexStorageFactory)),
+      Queue(std::move(Opts.OnProgress)),
       CommandsChanged(
           CDB.watch([&](const std::vector<std::string> &ChangedFiles) {
             enqueue(ChangedFiles);
           })) {
-  assert(ThreadPoolSize > 0 && "Thread pool size can't be zero.");
+  assert(Opts.ThreadPoolSize > 0 && "Thread pool size can't be zero.");
   assert(this->IndexStorageFactory && "Storage factory can not be null!");
-  for (unsigned I = 0; I < ThreadPoolSize; ++I) {
-    ThreadPool.runAsync("background-worker-" + llvm::Twine(I + 1), [this] {
-      WithContext Ctx(this->BackgroundContext.clone());
-      Queue.work([&] { Rebuilder.idle(); });
-    });
+  for (unsigned I = 0; I < Opts.ThreadPoolSize; ++I) {
+    ThreadPool.runAsync("background-worker-" + llvm::Twine(I + 1),
+                        [this, Ctx(Context::current().clone())]() mutable {
+                          WithContext BGContext(std::move(Ctx));
+                          Queue.work([&] { Rebuilder.idle(); });
+                        });
   }
 }
 
@@ -165,6 +125,11 @@ BackgroundQueue::Task BackgroundIndex::changedFilesTask(
     const std::vector<std::string> &ChangedFiles) {
   BackgroundQueue::Task T([this, ChangedFiles] {
     trace::Span Tracer("BackgroundIndexEnqueue");
+
+    llvm::Optional<WithContext> WithProvidedContext;
+    if (ContextProvider)
+      WithProvidedContext.emplace(ContextProvider(/*Path=*/""));
+
     // We're doing this asynchronously, because we'll read shards here too.
     log("Enqueueing {0} commands for indexing", ChangedFiles.size());
     SPAN_ATTACH(Tracer, "files", int64_t(ChangedFiles.size()));
@@ -175,8 +140,8 @@ BackgroundQueue::Task BackgroundIndex::changedFilesTask(
                  std::mt19937(std::random_device{}()));
     std::vector<BackgroundQueue::Task> Tasks;
     Tasks.reserve(NeedsReIndexing.size());
-    for (auto &Cmd : NeedsReIndexing)
-      Tasks.push_back(indexFileTask(std::move(Cmd)));
+    for (const auto &File : NeedsReIndexing)
+      Tasks.push_back(indexFileTask(std::move(File)));
     Queue.append(std::move(Tasks));
   });
 
@@ -190,17 +155,22 @@ static llvm::StringRef filenameWithoutExtension(llvm::StringRef Path) {
   return Path.drop_back(llvm::sys::path::extension(Path).size());
 }
 
-BackgroundQueue::Task
-BackgroundIndex::indexFileTask(tooling::CompileCommand Cmd) {
-  BackgroundQueue::Task T([this, Cmd] {
-    // We can't use llvm::StringRef here since we are going to
-    // move from Cmd during the call below.
-    const std::string FileName = Cmd.Filename;
-    if (auto Error = index(std::move(Cmd)))
-      elog("Indexing {0} failed: {1}", FileName, std::move(Error));
+BackgroundQueue::Task BackgroundIndex::indexFileTask(std::string Path) {
+  std::string Tag = filenameWithoutExtension(Path).str();
+  uint64_t Key = llvm::xxHash64(Path);
+  BackgroundQueue::Task T([this, Path(std::move(Path))] {
+    llvm::Optional<WithContext> WithProvidedContext;
+    if (ContextProvider)
+      WithProvidedContext.emplace(ContextProvider(Path));
+    auto Cmd = CDB.getCompileCommand(Path);
+    if (!Cmd)
+      return;
+    if (auto Error = index(std::move(*Cmd)))
+      elog("Indexing {0} failed: {1}", Path, std::move(Error));
   });
   T.QueuePri = IndexFile;
-  T.Tag = filenameWithoutExtension(Cmd.Filename);
+  T.Tag = std::move(Tag);
+  T.Key = Key;
   return T;
 }
 
@@ -216,108 +186,50 @@ void BackgroundIndex::update(
     llvm::StringRef MainFile, IndexFileIn Index,
     const llvm::StringMap<ShardVersion> &ShardVersionsSnapshot,
     bool HadErrors) {
-  // Partition symbols/references into files.
-  struct File {
-    llvm::DenseSet<const Symbol *> Symbols;
-    llvm::DenseSet<const Ref *> Refs;
-    llvm::DenseSet<const Relation *> Relations;
-    FileDigest Digest;
-  };
-  llvm::StringMap<File> Files;
-  URIToFileCache URICache(MainFile);
+  // Keys are URIs.
+  llvm::StringMap<std::pair<Path, FileDigest>> FilesToUpdate;
+  // Note that sources do not contain any information regarding missing headers,
+  // since we don't even know what absolute path they should fall in.
   for (const auto &IndexIt : *Index.Sources) {
     const auto &IGN = IndexIt.getValue();
-    // Note that sources do not contain any information regarding missing
-    // headers, since we don't even know what absolute path they should fall in.
-    const auto AbsPath = URICache.resolve(IGN.URI);
-    const auto DigestIt = ShardVersionsSnapshot.find(AbsPath);
-    // File has different contents, or indexing was successfull this time.
+    auto AbsPath = URI::resolve(IGN.URI, MainFile);
+    if (!AbsPath) {
+      elog("Failed to resolve URI: {0}", AbsPath.takeError());
+      continue;
+    }
+    const auto DigestIt = ShardVersionsSnapshot.find(*AbsPath);
+    // File has different contents, or indexing was successful this time.
     if (DigestIt == ShardVersionsSnapshot.end() ||
         DigestIt->getValue().Digest != IGN.Digest ||
         (DigestIt->getValue().HadErrors && !HadErrors))
-      Files.try_emplace(AbsPath).first->getValue().Digest = IGN.Digest;
-  }
-  // This map is used to figure out where to store relations.
-  llvm::DenseMap<SymbolID, File *> SymbolIDToFile;
-  for (const auto &Sym : *Index.Symbols) {
-    if (Sym.CanonicalDeclaration) {
-      auto DeclPath = URICache.resolve(Sym.CanonicalDeclaration.FileURI);
-      const auto FileIt = Files.find(DeclPath);
-      if (FileIt != Files.end()) {
-        FileIt->second.Symbols.insert(&Sym);
-        SymbolIDToFile[Sym.ID] = &FileIt->second;
-      }
-    }
-    // For symbols with different declaration and definition locations, we store
-    // the full symbol in both the header file and the implementation file, so
-    // that merging can tell the preferred symbols (from canonical headers) from
-    // other symbols (e.g. forward declarations).
-    if (Sym.Definition &&
-        Sym.Definition.FileURI != Sym.CanonicalDeclaration.FileURI) {
-      auto DefPath = URICache.resolve(Sym.Definition.FileURI);
-      const auto FileIt = Files.find(DefPath);
-      if (FileIt != Files.end())
-        FileIt->second.Symbols.insert(&Sym);
-    }
-  }
-  llvm::DenseMap<const Ref *, SymbolID> RefToIDs;
-  for (const auto &SymRefs : *Index.Refs) {
-    for (const auto &R : SymRefs.second) {
-      auto Path = URICache.resolve(R.Location.FileURI);
-      const auto FileIt = Files.find(Path);
-      if (FileIt != Files.end()) {
-        auto &F = FileIt->getValue();
-        RefToIDs[&R] = SymRefs.first;
-        F.Refs.insert(&R);
-      }
-    }
-  }
-  for (const auto &Rel : *Index.Relations) {
-    const auto FileIt = SymbolIDToFile.find(Rel.Subject);
-    if (FileIt != SymbolIDToFile.end())
-      FileIt->second->Relations.insert(&Rel);
+      FilesToUpdate[IGN.URI] = {std::move(*AbsPath), IGN.Digest};
   }
 
+  // Shard slabs into files.
+  FileShardedIndex ShardedIndex(std::move(Index));
+
   // Build and store new slabs for each updated file.
-  for (const auto &FileIt : Files) {
-    llvm::StringRef Path = FileIt.getKey();
-    SymbolSlab::Builder Syms;
-    RefSlab::Builder Refs;
-    RelationSlab::Builder Relations;
-    for (const auto *S : FileIt.second.Symbols)
-      Syms.insert(*S);
-    for (const auto *R : FileIt.second.Refs)
-      Refs.insert(RefToIDs[R], *R);
-    for (const auto *Rel : FileIt.second.Relations)
-      Relations.insert(*Rel);
-    auto SS = std::make_unique<SymbolSlab>(std::move(Syms).build());
-    auto RS = std::make_unique<RefSlab>(std::move(Refs).build());
-    auto RelS = std::make_unique<RelationSlab>(std::move(Relations).build());
-    auto IG = std::make_unique<IncludeGraph>(
-        getSubGraph(URI::create(Path), Index.Sources.getValue()));
+  for (const auto &FileIt : FilesToUpdate) {
+    auto Uri = FileIt.first();
+    auto IF = ShardedIndex.getShard(Uri);
+    assert(IF && "no shard for file in Index.Sources?");
+    PathRef Path = FileIt.getValue().first;
+
+    // Only store command line hash for main files of the TU, since our
+    // current model keeps only one version of a header file.
+    if (Path != MainFile)
+      IF->Cmd.reset();
 
     // We need to store shards before updating the index, since the latter
     // consumes slabs.
     // FIXME: Also skip serializing the shard if it is already up-to-date.
-    BackgroundIndexStorage *IndexStorage = IndexStorageFactory(Path);
-    IndexFileOut Shard;
-    Shard.Symbols = SS.get();
-    Shard.Refs = RS.get();
-    Shard.Relations = RelS.get();
-    Shard.Sources = IG.get();
-
-    // Only store command line hash for main files of the TU, since our
-    // current model keeps only one version of a header file.
-    if (Path == MainFile)
-      Shard.Cmd = Index.Cmd.getPointer();
-
-    if (auto Error = IndexStorage->storeShard(Path, Shard))
+    if (auto Error = IndexStorageFactory(Path)->storeShard(Path, *IF))
       elog("Failed to write background-index shard for file {0}: {1}", Path,
            std::move(Error));
 
     {
       std::lock_guard<std::mutex> Lock(ShardVersionsMu);
-      auto Hash = FileIt.second.Digest;
+      const auto &Hash = FileIt.getValue().second;
       auto DigestIt = ShardVersions.try_emplace(Path);
       ShardVersion &SV = DigestIt.first->second;
       // Skip if file is already up to date, unless previous index was broken
@@ -330,8 +242,11 @@ void BackgroundIndex::update(
       // This can override a newer version that is added in another thread, if
       // this thread sees the older version but finishes later. This should be
       // rare in practice.
-      IndexedSymbols.update(Path, std::move(SS), std::move(RS), std::move(RelS),
-                            Path == MainFile);
+      IndexedSymbols.update(
+          Uri, std::make_unique<SymbolSlab>(std::move(*IF->Symbols)),
+          std::make_unique<RefSlab>(std::move(*IF->Refs)),
+          std::make_unique<RelationSlab>(std::move(*IF->Relations)),
+          Path == MainFile);
     }
   }
 }
@@ -341,7 +256,7 @@ llvm::Error BackgroundIndex::index(tooling::CompileCommand Cmd) {
   SPAN_ATTACH(Tracer, "file", Cmd.Filename);
   auto AbsolutePath = getAbsolutePath(Cmd);
 
-  auto FS = FSProvider.getFileSystem();
+  auto FS = TFS.view(Cmd.Directory);
   auto Buf = FS->getBufferForFile(AbsolutePath);
   if (!Buf)
     return llvm::errorCodeToError(Buf.getError());
@@ -356,19 +271,18 @@ llvm::Error BackgroundIndex::index(tooling::CompileCommand Cmd) {
 
   vlog("Indexing {0} (digest:={1})", Cmd.Filename, llvm::toHex(Hash));
   ParseInputs Inputs;
-  Inputs.FS = std::move(FS);
-  Inputs.FS->setCurrentWorkingDirectory(Cmd.Directory);
+  Inputs.TFS = &TFS;
   Inputs.CompileCommand = std::move(Cmd);
   IgnoreDiagnostics IgnoreDiags;
   auto CI = buildCompilerInvocation(Inputs, IgnoreDiags);
   if (!CI)
-    return llvm::createStringError(llvm::inconvertibleErrorCode(),
-                                   "Couldn't build compiler invocation");
-  auto Clang = prepareCompilerInstance(std::move(CI), /*Preamble=*/nullptr,
-                                       std::move(*Buf), Inputs.FS, IgnoreDiags);
+    return error("Couldn't build compiler invocation");
+
+  auto Clang =
+      prepareCompilerInstance(std::move(CI), /*Preamble=*/nullptr,
+                              std::move(*Buf), std::move(FS), IgnoreDiags);
   if (!Clang)
-    return llvm::createStringError(llvm::inconvertibleErrorCode(),
-                                   "Couldn't build compiler instance");
+    return error("Couldn't build compiler instance");
 
   SymbolCollector::Options IndexOpts;
   // Creates a filter to not collect index results from files with unchanged
@@ -390,6 +304,7 @@ llvm::Error BackgroundIndex::index(tooling::CompileCommand Cmd) {
       return false; // Skip files that haven't changed, without errors.
     return true;
   };
+  IndexOpts.CollectMainFileRefs = true;
 
   IndexFileIn Index;
   auto Action = createStaticIndexingAction(
@@ -405,8 +320,7 @@ llvm::Error BackgroundIndex::index(tooling::CompileCommand Cmd) {
 
   const FrontendInputFile &Input = Clang->getFrontendOpts().Inputs.front();
   if (!Action->BeginSourceFile(*Clang, Input))
-    return llvm::createStringError(llvm::inconvertibleErrorCode(),
-                                   "BeginSourceFile() failed");
+    return error("BeginSourceFile() failed");
   if (llvm::Error Err = Action->Execute())
     return Err;
 
@@ -439,10 +353,16 @@ llvm::Error BackgroundIndex::index(tooling::CompileCommand Cmd) {
 // Restores shards for \p MainFiles from index storage. Then checks staleness of
 // those shards and returns a list of TUs that needs to be indexed to update
 // staleness.
-std::vector<tooling::CompileCommand>
+std::vector<std::string>
 BackgroundIndex::loadProject(std::vector<std::string> MainFiles) {
-  std::vector<tooling::CompileCommand> NeedsReIndexing;
-
+  // Drop files where background indexing is disabled in config.
+  if (ContextProvider)
+    llvm::erase_if(MainFiles, [&](const std::string &TU) {
+      // Load the config for each TU, as indexing may be selectively enabled.
+      WithContext WithProvidedContext(ContextProvider(TU));
+      return Config::current().Index.Background ==
+             Config::BackgroundPolicy::Skip;
+    });
   Rebuilder.startLoading();
   // Load shards for all of the mainfiles.
   const std::vector<LoadedShard> Result =
@@ -470,14 +390,15 @@ BackgroundIndex::loadProject(std::vector<std::string> MainFiles) {
       SV.HadErrors = LS.HadErrors;
       ++LoadedShards;
 
-      IndexedSymbols.update(LS.AbsolutePath, std::move(SS), std::move(RS),
-                            std::move(RelS), LS.CountReferences);
+      IndexedSymbols.update(URI::create(LS.AbsolutePath).toString(),
+                            std::move(SS), std::move(RS), std::move(RelS),
+                            LS.CountReferences);
     }
   }
   Rebuilder.loadedShard(LoadedShards);
   Rebuilder.doneLoading();
 
-  auto FS = FSProvider.getFileSystem();
+  auto FS = TFS.view(/*CWD=*/llvm::None);
   llvm::DenseSet<PathRef> TUsToIndex;
   // We'll accept data from stale shards, but ensure the files get reindexed
   // soon.
@@ -495,15 +416,13 @@ BackgroundIndex::loadProject(std::vector<std::string> MainFiles) {
     TUsToIndex.insert(TUForFile);
   }
 
-  for (PathRef TU : TUsToIndex) {
-    auto Cmd = CDB.getCompileCommand(TU);
-    if (!Cmd)
-      continue;
-    NeedsReIndexing.emplace_back(std::move(*Cmd));
-  }
-
-  return NeedsReIndexing;
+  return {TUsToIndex.begin(), TUsToIndex.end()};
 }
 
+void BackgroundIndex::profile(MemoryTree &MT) const {
+  IndexedSymbols.profile(MT.child("slabs"));
+  // We don't want to mix memory used by index and symbols, so call base class.
+  MT.child("index").addUsage(SwapIndex::estimateMemoryUsage());
+}
 } // namespace clangd
 } // namespace clang
