@@ -1,18 +1,16 @@
 //! Implements "Stacked Borrows".  See <https://github.com/rust-lang/unsafe-code-guidelines/blob/master/wip/stacked-borrows.md>
 //! for further information.
 
+use log::trace;
 use std::cell::RefCell;
 use std::fmt;
 use std::num::NonZeroU64;
-use std::rc::Rc;
-
-use log::trace;
 
 use rustc_data_structures::fx::{FxHashMap, FxHashSet};
-use rustc_middle::mir::RetagKind;
-use rustc_middle::ty;
-use rustc_target::abi::{Align, LayoutOf, Size};
 use rustc_hir::Mutability;
+use rustc_middle::mir::RetagKind;
+use rustc_middle::ty::{self, layout::LayoutOf};
+use rustc_target::abi::Size;
 
 use crate::*;
 
@@ -22,16 +20,16 @@ pub type AllocExtra = Stacks;
 
 /// Tracking pointer provenance
 #[derive(Copy, Clone, Hash, PartialEq, Eq)]
-pub enum Tag {
+pub enum SbTag {
     Tagged(PtrId),
     Untagged,
 }
 
-impl fmt::Debug for Tag {
+impl fmt::Debug for SbTag {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         match self {
-            Tag::Tagged(id) => write!(f, "<{}>", id),
-            Tag::Untagged => write!(f, "<untagged>"),
+            SbTag::Tagged(id) => write!(f, "<{}>", id),
+            SbTag::Untagged => write!(f, "<untagged>"),
         }
     }
 }
@@ -56,7 +54,7 @@ pub struct Item {
     /// The permission this item grants.
     perm: Permission,
     /// The pointers the permission is granted to.
-    tag: Tag,
+    tag: SbTag,
     /// An optional protector, ensuring the item cannot get popped until `CallId` is over.
     protector: Option<CallId>,
 }
@@ -87,8 +85,6 @@ pub struct Stack {
 pub struct Stacks {
     // Even reading memory can have effects on the stack, so we need a `RefCell` here.
     stacks: RefCell<RangeMap<Stack>>,
-    // Pointer to global state
-    global: MemoryExtra,
 }
 
 /// Extra global state, available to the memory access hooks.
@@ -99,7 +95,7 @@ pub struct GlobalState {
     /// Table storing the "base" tag for each allocation.
     /// The base tag is the one used for the initial pointer.
     /// We need this in a separate table to handle cyclic statics.
-    base_ptr_ids: FxHashMap<AllocId, Tag>,
+    base_ptr_ids: FxHashMap<AllocId, SbTag>,
     /// Next unused call ID (for protectors).
     next_call_id: CallId,
     /// Those call IDs corresponding to functions that are still running.
@@ -108,9 +104,11 @@ pub struct GlobalState {
     tracked_pointer_tag: Option<PtrId>,
     /// The call id to trace
     tracked_call_id: Option<CallId>,
+    /// Whether to track raw pointers.
+    track_raw: bool,
 }
 /// Memory extra state gives us interior mutable access to the global state.
-pub type MemoryExtra = Rc<RefCell<GlobalState>>;
+pub type MemoryExtra = RefCell<GlobalState>;
 
 /// Indicates which kind of access is being performed.
 #[derive(Copy, Clone, Hash, PartialEq, Eq)]
@@ -155,7 +153,11 @@ impl fmt::Display for RefKind {
 
 /// Utilities for initialization and ID generation
 impl GlobalState {
-    pub fn new(tracked_pointer_tag: Option<PtrId>, tracked_call_id: Option<CallId>) -> Self {
+    pub fn new(
+        tracked_pointer_tag: Option<PtrId>,
+        tracked_call_id: Option<CallId>,
+        track_raw: bool,
+    ) -> Self {
         GlobalState {
             next_ptr_id: NonZeroU64::new(1).unwrap(),
             base_ptr_ids: FxHashMap::default(),
@@ -163,11 +165,15 @@ impl GlobalState {
             active_calls: FxHashSet::default(),
             tracked_pointer_tag,
             tracked_call_id,
+            track_raw,
         }
     }
 
     fn new_ptr(&mut self) -> PtrId {
         let id = self.next_ptr_id;
+        if Some(id) == self.tracked_pointer_tag {
+            register_diagnostic(NonHaltingDiagnostic::CreatedPointerTag(id));
+        }
         self.next_ptr_id = NonZeroU64::new(id.get() + 1).unwrap();
         id
     }
@@ -191,21 +197,31 @@ impl GlobalState {
         self.active_calls.contains(&id)
     }
 
-    pub fn global_base_ptr(&mut self, id: AllocId) -> Tag {
+    pub fn base_tag(&mut self, id: AllocId) -> SbTag {
         self.base_ptr_ids.get(&id).copied().unwrap_or_else(|| {
-            let tag = Tag::Tagged(self.new_ptr());
+            let tag = SbTag::Tagged(self.new_ptr());
             trace!("New allocation {:?} has base tag {:?}", id, tag);
-            self.base_ptr_ids.insert(id, tag).unwrap_none();
+            self.base_ptr_ids.try_insert(id, tag).unwrap();
             tag
         })
     }
+
+    pub fn base_tag_untagged(&mut self, id: AllocId) -> SbTag {
+        trace!("New allocation {:?} has no base tag (untagged)", id);
+        let tag = SbTag::Untagged;
+        // This must only be done on new allocations.
+        self.base_ptr_ids.try_insert(id, tag).unwrap();
+        tag
+    }
 }
 
 /// Error reporting
 fn err_sb_ub(msg: String) -> InterpError<'static> {
     err_machine_stop!(TerminationInfo::ExperimentalUb {
         msg,
-        url: format!("https://github.com/rust-lang/unsafe-code-guidelines/blob/master/wip/stacked-borrows.md"),
+        url: format!(
+            "https://github.com/rust-lang/unsafe-code-guidelines/blob/master/wip/stacked-borrows.md"
+        ),
     })
 }
 
@@ -239,7 +255,7 @@ impl Permission {
 impl<'tcx> Stack {
     /// Find the item granting the given kind of access to the given tag, and return where
     /// it is on the stack.
-    fn find_granting(&self, access: AccessKind, tag: Tag) -> Option<usize> {
+    fn find_granting(&self, access: AccessKind, tag: SbTag) -> Option<usize> {
         self.borrows
             .iter()
             .enumerate() // we also need to know *where* in the stack
@@ -280,8 +296,12 @@ impl<'tcx> Stack {
     }
 
     /// Check if the given item is protected.
-    fn check_protector(item: &Item, tag: Option<Tag>, global: &GlobalState) -> InterpResult<'tcx> {
-        if let Tag::Tagged(id) = item.tag {
+    fn check_protector(
+        item: &Item,
+        tag: Option<SbTag>,
+        global: &GlobalState,
+    ) -> InterpResult<'tcx> {
+        if let SbTag::Tagged(id) = item.tag {
             if Some(id) == global.tracked_pointer_tag {
                 register_diagnostic(NonHaltingDiagnostic::PoppedPointerTag(item.clone()));
             }
@@ -294,10 +314,7 @@ impl<'tcx> Stack {
                         tag, item
                     )))?
                 } else {
-                    Err(err_sb_ub(format!(
-                        "deallocating while item is protected: {:?}",
-                        item
-                    )))?
+                    Err(err_sb_ub(format!("deallocating while item is protected: {:?}", item)))?
                 }
             }
         }
@@ -306,14 +323,20 @@ impl<'tcx> Stack {
 
     /// Test if a memory `access` using pointer tagged `tag` is granted.
     /// If yes, return the index of the item that granted it.
-    fn access(&mut self, access: AccessKind, tag: Tag, global: &GlobalState) -> InterpResult<'tcx> {
+    fn access(
+        &mut self,
+        access: AccessKind,
+        tag: SbTag,
+        dbg_ptr: Pointer<AllocId>, // just for debug printing amd error messages
+        global: &GlobalState,
+    ) -> InterpResult<'tcx> {
         // Two main steps: Find granting item, remove incompatible items above.
 
         // Step 1: Find granting item.
         let granting_idx = self.find_granting(access, tag).ok_or_else(|| {
             err_sb_ub(format!(
-                "no item granting {} to tag {:?} found in borrow stack.",
-                access, tag
+                "no item granting {} to tag {:?} at {:?} found in borrow stack.",
+                access, tag, dbg_ptr,
             ))
         })?;
 
@@ -352,12 +375,17 @@ impl<'tcx> Stack {
 
     /// Deallocate a location: Like a write access, but also there must be no
     /// active protectors at all because we will remove all items.
-    fn dealloc(&mut self, tag: Tag, global: &GlobalState) -> InterpResult<'tcx> {
+    fn dealloc(
+        &mut self,
+        tag: SbTag,
+        dbg_ptr: Pointer<AllocId>, // just for debug printing amd error messages
+        global: &GlobalState,
+    ) -> InterpResult<'tcx> {
         // Step 1: Find granting item.
         self.find_granting(AccessKind::Write, tag).ok_or_else(|| {
             err_sb_ub(format!(
-                "no item granting write access for deallocation to tag {:?} found in borrow stack",
-                tag,
+                "no item granting write access for deallocation to tag {:?} at {:?} found in borrow stack",
+                tag, dbg_ptr,
             ))
         })?;
 
@@ -369,11 +397,17 @@ impl<'tcx> Stack {
         Ok(())
     }
 
-    /// Derived a new pointer from one with the given tag.
+    /// Derive a new pointer from one with the given tag.
     /// `weak` controls whether this operation is weak or strong: weak granting does not act as
     /// an access, and they add the new item directly on top of the one it is derived
     /// from instead of all the way at the top of the stack.
-    fn grant(&mut self, derived_from: Tag, new: Item, global: &GlobalState) -> InterpResult<'tcx> {
+    fn grant(
+        &mut self,
+        derived_from: SbTag,
+        new: Item,
+        dbg_ptr: Pointer<AllocId>,
+        global: &GlobalState,
+    ) -> InterpResult<'tcx> {
         // Figure out which access `perm` corresponds to.
         let access =
             if new.perm.grants(AccessKind::Write) { AccessKind::Write } else { AccessKind::Read };
@@ -381,8 +415,8 @@ impl<'tcx> Stack {
         // We use that to determine where to put the new item.
         let granting_idx = self.find_granting(access, derived_from)
             .ok_or_else(|| err_sb_ub(format!(
-                "trying to reborrow for {:?}, but parent tag {:?} does not have an appropriate item in the borrow stack",
-                new.perm, derived_from,
+                "trying to reborrow for {:?} at {:?}, but parent tag {:?} does not have an appropriate item in the borrow stack",
+                new.perm, dbg_ptr, derived_from,
             )))?;
 
         // Compute where to put the new item.
@@ -402,7 +436,7 @@ impl<'tcx> Stack {
             // A "safe" reborrow for a pointer that actually expects some aliasing guarantees.
             // Here, creating a reference actually counts as an access.
             // This ensures F2b for `Unique`, by removing offending `SharedReadOnly`.
-            self.access(access, derived_from, global)?;
+            self.access(access, derived_from, dbg_ptr, global)?;
 
             // We insert "as far up as possible": We know only compatible items are remaining
             // on top of `derived_from`, and we want the new item at the top so that we
@@ -428,24 +462,35 @@ impl<'tcx> Stack {
 /// Map per-stack operations to higher-level per-location-range operations.
 impl<'tcx> Stacks {
     /// Creates new stack with initial tag.
-    fn new(size: Size, perm: Permission, tag: Tag, extra: MemoryExtra) -> Self {
+    fn new(size: Size, perm: Permission, tag: SbTag) -> Self {
         let item = Item { perm, tag, protector: None };
         let stack = Stack { borrows: vec![item] };
 
-        Stacks { stacks: RefCell::new(RangeMap::new(size, stack)), global: extra }
+        Stacks { stacks: RefCell::new(RangeMap::new(size, stack)) }
     }
 
     /// Call `f` on every stack in the range.
     fn for_each(
         &self,
-        ptr: Pointer<Tag>,
-        size: Size,
-        f: impl Fn(&mut Stack, &GlobalState) -> InterpResult<'tcx>,
+        range: AllocRange,
+        f: impl Fn(Size, &mut Stack) -> InterpResult<'tcx>,
     ) -> InterpResult<'tcx> {
-        let global = self.global.borrow();
         let mut stacks = self.stacks.borrow_mut();
-        for stack in stacks.iter_mut(ptr.offset, size) {
-            f(stack, &*global)?;
+        for (offset, stack) in stacks.iter_mut(range.start, range.size) {
+            f(offset, stack)?;
+        }
+        Ok(())
+    }
+
+    /// Call `f` on every stack in the range.
+    fn for_each_mut(
+        &mut self,
+        range: AllocRange,
+        f: impl Fn(Size, &mut Stack) -> InterpResult<'tcx>,
+    ) -> InterpResult<'tcx> {
+        let stacks = self.stacks.get_mut();
+        for (offset, stack) in stacks.iter_mut(range.start, range.size) {
+            f(offset, stack)?;
         }
         Ok(())
     }
@@ -456,57 +501,97 @@ impl Stacks {
     pub fn new_allocation(
         id: AllocId,
         size: Size,
-        extra: MemoryExtra,
+        extra: &MemoryExtra,
         kind: MemoryKind<MiriMemoryKind>,
-    ) -> (Self, Tag) {
-        let (tag, perm) = match kind {
+    ) -> Self {
+        let mut extra = extra.borrow_mut();
+        let (base_tag, perm) = match kind {
             // New unique borrow. This tag is not accessible by the program,
             // so it will only ever be used when using the local directly (i.e.,
             // not through a pointer). That is, whenever we directly write to a local, this will pop
             // everything else off the stack, invalidating all previous pointers,
             // and in particular, *all* raw pointers.
-            MemoryKind::Stack => (Tag::Tagged(extra.borrow_mut().new_ptr()), Permission::Unique),
+            MemoryKind::Stack => (extra.base_tag(id), Permission::Unique),
             // `Global` memory can be referenced by global pointers from `tcx`.
             // Thus we call `global_base_ptr` such that the global pointers get the same tag
             // as what we use here.
-            // `ExternGlobal` is used for extern statics, and thus must also be listed here.
-            // `Env` we list because we can get away with precise tracking there.
+            // `ExternStatic` is used for extern statics, so the same reasoning applies.
+            // The others are various forms of machine-managed special global memory, and we can get
+            // away with precise tracking there.
             // The base pointer is not unique, so the base permission is `SharedReadWrite`.
-            MemoryKind::Machine(MiriMemoryKind::Global | MiriMemoryKind::ExternGlobal | MiriMemoryKind::Env) =>
-                (extra.borrow_mut().global_base_ptr(id), Permission::SharedReadWrite),
-            // Everything else we handle entirely untagged for now.
-            // FIXME: experiment with more precise tracking.
-            _ => (Tag::Untagged, Permission::SharedReadWrite),
+            MemoryKind::CallerLocation
+            | MemoryKind::Machine(
+                MiriMemoryKind::Global
+                | MiriMemoryKind::ExternStatic
+                | MiriMemoryKind::Tls
+                | MiriMemoryKind::Env
+                | MiriMemoryKind::Machine,
+            ) => (extra.base_tag(id), Permission::SharedReadWrite),
+            // Heap allocations we only track precisely when raw pointers are tagged, for now.
+            MemoryKind::Machine(
+                MiriMemoryKind::Rust | MiriMemoryKind::C | MiriMemoryKind::WinHeap,
+            ) => {
+                let tag =
+                    if extra.track_raw { extra.base_tag(id) } else { extra.base_tag_untagged(id) };
+                (tag, Permission::SharedReadWrite)
+            }
         };
-        (Stacks::new(size, perm, tag, extra), tag)
+        Stacks::new(size, perm, base_tag)
     }
 
     #[inline(always)]
-    pub fn memory_read<'tcx>(&self, ptr: Pointer<Tag>, size: Size) -> InterpResult<'tcx> {
-        trace!("read access with tag {:?}: {:?}, size {}", ptr.tag, ptr.erase_tag(), size.bytes());
-        self.for_each(ptr, size, |stack, global| {
-            stack.access(AccessKind::Read, ptr.tag, global)?;
-            Ok(())
+    pub fn memory_read<'tcx>(
+        &self,
+        alloc_id: AllocId,
+        tag: SbTag,
+        range: AllocRange,
+        extra: &MemoryExtra,
+    ) -> InterpResult<'tcx> {
+        trace!(
+            "read access with tag {:?}: {:?}, size {}",
+            tag,
+            Pointer::new(alloc_id, range.start),
+            range.size.bytes()
+        );
+        let global = &*extra.borrow();
+        self.for_each(range, move |offset, stack| {
+            stack.access(AccessKind::Read, tag, Pointer::new(alloc_id, offset), global)
         })
     }
 
     #[inline(always)]
-    pub fn memory_written<'tcx>(&mut self, ptr: Pointer<Tag>, size: Size) -> InterpResult<'tcx> {
-        trace!("write access with tag {:?}: {:?}, size {}", ptr.tag, ptr.erase_tag(), size.bytes());
-        self.for_each(ptr, size, |stack, global| {
-            stack.access(AccessKind::Write, ptr.tag, global)?;
-            Ok(())
+    pub fn memory_written<'tcx>(
+        &mut self,
+        alloc_id: AllocId,
+        tag: SbTag,
+        range: AllocRange,
+        extra: &mut MemoryExtra,
+    ) -> InterpResult<'tcx> {
+        trace!(
+            "write access with tag {:?}: {:?}, size {}",
+            tag,
+            Pointer::new(alloc_id, range.start),
+            range.size.bytes()
+        );
+        let global = extra.get_mut();
+        self.for_each_mut(range, move |offset, stack| {
+            stack.access(AccessKind::Write, tag, Pointer::new(alloc_id, offset), global)
         })
     }
 
     #[inline(always)]
     pub fn memory_deallocated<'tcx>(
         &mut self,
-        ptr: Pointer<Tag>,
-        size: Size,
+        alloc_id: AllocId,
+        tag: SbTag,
+        range: AllocRange,
+        extra: &mut MemoryExtra,
     ) -> InterpResult<'tcx> {
-        trace!("deallocation with tag {:?}: {:?}, size {}", ptr.tag, ptr.erase_tag(), size.bytes());
-        self.for_each(ptr, size, |stack, global| stack.dealloc(ptr.tag, global))
+        trace!("deallocation with tag {:?}: {:?}, size {}", tag, alloc_id, range.size.bytes());
+        let global = extra.get_mut();
+        self.for_each_mut(range, move |offset, stack| {
+            stack.dealloc(tag, Pointer::new(alloc_id, offset), global)
+        })
     }
 }
 
@@ -516,29 +601,51 @@ impl<'mir, 'tcx: 'mir> EvalContextPrivExt<'mir, 'tcx> for crate::MiriEvalContext
 trait EvalContextPrivExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx> {
     fn reborrow(
         &mut self,
-        place: MPlaceTy<'tcx, Tag>,
+        place: &MPlaceTy<'tcx, Tag>,
         size: Size,
         kind: RefKind,
-        new_tag: Tag,
+        new_tag: SbTag,
         protect: bool,
     ) -> InterpResult<'tcx> {
         let this = self.eval_context_mut();
+        if size == Size::ZERO {
+            // Nothing to do for zero-sized accesses.
+            trace!(
+                "reborrow of size 0: {} reference {:?} derived from {:?} (pointee {})",
+                kind,
+                new_tag,
+                place.ptr,
+                place.layout.ty,
+            );
+            return Ok(());
+        }
+        let (alloc_id, base_offset, ptr) = this.memory.ptr_get_alloc(place.ptr)?;
+        let orig_tag = ptr.provenance.sb;
+
+        // Ensure we bail out if the pointer goes out-of-bounds (see miri#1050).
+        let (alloc_size, _) =
+            this.memory.get_size_and_align(alloc_id, AllocCheck::Dereferenceable)?;
+        if base_offset + size > alloc_size {
+            throw_ub!(PointerOutOfBounds {
+                alloc_id,
+                alloc_size,
+                ptr_offset: this.machine_usize_to_isize(base_offset.bytes()),
+                ptr_size: size,
+                msg: CheckInAllocMsg::InboundsTest
+            });
+        }
+
         let protector = if protect { Some(this.frame().extra.call_id) } else { None };
-        let ptr = place.ptr.assert_ptr();
         trace!(
             "reborrow: {} reference {:?} derived from {:?} (pointee {}): {:?}, size {}",
             kind,
             new_tag,
-            ptr.tag,
+            orig_tag,
             place.layout.ty,
-            ptr.erase_tag(),
+            Pointer::new(alloc_id, base_offset),
             size.bytes()
         );
 
-        // Get the allocation. It might not be mutable, so we cannot use `get_mut`.
-        let extra = &this.memory.get_raw(ptr.alloc_id)?.extra;
-        let stacked_borrows =
-            extra.stacked_borrows.as_ref().expect("we should have Stacked Borrows data");
         // Update the stacks.
         // Make sure that raw pointers and mutable shared references are reborrowed "weak":
         // There could be existing unique pointers reborrowed from them that should remain valid!
@@ -550,7 +657,15 @@ trait EvalContextPrivExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
                 // Shared references and *const are a whole different kind of game, the
                 // permission is not uniform across the entire range!
                 // We need a frozen-sensitive reborrow.
-                return this.visit_freeze_sensitive(place, size, |cur_ptr, size, frozen| {
+                // We have to use shared references to alloc/memory_extra here since
+                // `visit_freeze_sensitive` needs to access the global state.
+                let extra = this.memory.get_alloc_extra(alloc_id)?;
+                let stacked_borrows =
+                    extra.stacked_borrows.as_ref().expect("we should have Stacked Borrows data");
+                let global = this.memory.extra.stacked_borrows.as_ref().unwrap().borrow();
+                this.visit_freeze_sensitive(place, size, |mut range, frozen| {
+                    // Adjust range.
+                    range.start += base_offset;
                     // We are only ever `SharedReadOnly` inside the frozen bits.
                     let perm = if frozen {
                         Permission::SharedReadOnly
@@ -558,77 +673,84 @@ trait EvalContextPrivExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
                         Permission::SharedReadWrite
                     };
                     let item = Item { perm, tag: new_tag, protector };
-                    stacked_borrows.for_each(cur_ptr, size, |stack, global| {
-                        stack.grant(cur_ptr.tag, item, global)
+                    stacked_borrows.for_each(range, |offset, stack| {
+                        stack.grant(orig_tag, item, Pointer::new(alloc_id, offset), &*global)
                     })
-                });
+                })?;
+                return Ok(());
             }
         };
+        // Here we can avoid `borrow()` calls because we have mutable references.
+        // Note that this asserts that the allocation is mutable -- but since we are creating a
+        // mutable pointer, that seems reasonable.
+        let (alloc_extra, memory_extra) = this.memory.get_alloc_extra_mut(alloc_id)?;
+        let stacked_borrows =
+            alloc_extra.stacked_borrows.as_mut().expect("we should have Stacked Borrows data");
+        let global = memory_extra.stacked_borrows.as_mut().unwrap().get_mut();
         let item = Item { perm, tag: new_tag, protector };
-        stacked_borrows.for_each(ptr, size, |stack, global| stack.grant(ptr.tag, item, global))
+        stacked_borrows.for_each_mut(alloc_range(base_offset, size), |offset, stack| {
+            stack.grant(orig_tag, item, Pointer::new(alloc_id, offset), global)
+        })?;
+        Ok(())
     }
 
     /// Retags an indidual pointer, returning the retagged version.
     /// `mutbl` can be `None` to make this a raw pointer.
     fn retag_reference(
         &mut self,
-        val: ImmTy<'tcx, Tag>,
+        val: &ImmTy<'tcx, Tag>,
         kind: RefKind,
         protect: bool,
     ) -> InterpResult<'tcx, ImmTy<'tcx, Tag>> {
         let this = self.eval_context_mut();
         // We want a place for where the ptr *points to*, so we get one.
         let place = this.ref_to_mplace(val)?;
-        let size = this
-            .size_and_align_of_mplace(place)?
-            .map(|(size, _)| size)
-            .unwrap_or_else(|| place.layout.size);
-        // `reborrow` relies on getting a `Pointer` and everything being in-bounds,
-        // so let's ensure that. However, we do not care about alignment.
-        // We can see dangling ptrs in here e.g. after a Box's `Unique` was
-        // updated using "self.0 = ..." (can happen in Box::from_raw) so we cannot ICE; see miri#1050.
-        let place = this.mplace_access_checked(place, Some(Align::from_bytes(1).unwrap()))?;
-        // Nothing to do for ZSTs.
-        if size == Size::ZERO {
-            return Ok(val);
-        }
+        let size = this.size_and_align_of_mplace(&place)?.map(|(size, _)| size);
+        // FIXME: If we cannot determine the size (because the unsized tail is an `extern type`),
+        // bail out -- we cannot reasonably figure out which memory range to reborrow.
+        // See https://github.com/rust-lang/unsafe-code-guidelines/issues/276.
+        let size = match size {
+            Some(size) => size,
+            None => return Ok(*val),
+        };
 
         // Compute new borrow.
-        let new_tag = match kind {
-            // Give up tracking for raw pointers.
-            // FIXME: Experiment with more precise tracking. Blocked on `&raw`
-            // because `Rc::into_raw` currently creates intermediate references,
-            // breaking `Rc::from_raw`.
-            RefKind::Raw { .. } => Tag::Untagged,
-            // All other pointesr are properly tracked.
-            _ => Tag::Tagged(
-                this.memory.extra.stacked_borrows.as_ref().unwrap().borrow_mut().new_ptr(),
-            ),
+        let new_tag = {
+            let mem_extra = this.memory.extra.stacked_borrows.as_mut().unwrap().get_mut();
+            match kind {
+                // Give up tracking for raw pointers.
+                RefKind::Raw { .. } if !mem_extra.track_raw => SbTag::Untagged,
+                // All other pointers are properly tracked.
+                _ => SbTag::Tagged(mem_extra.new_ptr()),
+            }
         };
 
         // Reborrow.
-        this.reborrow(place, size, kind, new_tag, protect)?;
-        let new_place = place.replace_tag(new_tag);
+        this.reborrow(&place, size, kind, new_tag, protect)?;
+
+        // Adjust pointer.
+        let new_place = place.map_provenance(|p| p.map(|t| Tag { sb: new_tag, ..t }));
 
         // Return new pointer.
-        Ok(ImmTy::from_immediate(new_place.to_ref(), val.layout))
+        Ok(ImmTy::from_immediate(new_place.to_ref(this), val.layout))
     }
 }
 
 impl<'mir, 'tcx: 'mir> EvalContextExt<'mir, 'tcx> for crate::MiriEvalContext<'mir, 'tcx> {}
 pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx> {
-    fn retag(&mut self, kind: RetagKind, place: PlaceTy<'tcx, Tag>) -> InterpResult<'tcx> {
+    fn retag(&mut self, kind: RetagKind, place: &PlaceTy<'tcx, Tag>) -> InterpResult<'tcx> {
         let this = self.eval_context_mut();
         // Determine mutability and whether to add a protector.
         // Cannot use `builtin_deref` because that reports *immutable* for `Box`,
         // making it useless.
         fn qualify(ty: ty::Ty<'_>, kind: RetagKind) -> Option<(RefKind, bool)> {
-            match ty.kind {
+            match ty.kind() {
                 // References are simple.
-                ty::Ref(_, _, Mutability::Mut) => Some((
-                    RefKind::Unique { two_phase: kind == RetagKind::TwoPhase },
-                    kind == RetagKind::FnEntry,
-                )),
+                ty::Ref(_, _, Mutability::Mut) =>
+                    Some((
+                        RefKind::Unique { two_phase: kind == RetagKind::TwoPhase },
+                        kind == RetagKind::FnEntry,
+                    )),
                 ty::Ref(_, _, Mutability::Not) =>
                     Some((RefKind::Shared, kind == RetagKind::FnEntry)),
                 // Raw pointers need to be enabled.
@@ -646,8 +768,8 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
         // but might also cost us optimization and analyses. We will have to experiment more with this.
         if let Some((mutbl, protector)) = qualify(place.layout.ty, kind) {
             // Fast path.
-            let val = this.read_immediate(this.place_to_op(place)?)?;
-            let val = this.retag_reference(val, mutbl, protector)?;
+            let val = this.read_immediate(&this.place_to_op(place)?)?;
+            let val = this.retag_reference(&val, mutbl, protector)?;
             this.write_immediate(*val, place)?;
         }
 
@@ -656,7 +778,7 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
 
     /// After a stack frame got pushed, retag the return place so that we are sure
     /// it does not alias with anything.
-    /// 
+    ///
     /// This is a HACK because there is nothing in MIR that would make the retag
     /// explicit. Also see https://github.com/rust-lang/rust/issues/71117.
     fn retag_return_place(&mut self) -> InterpResult<'tcx> {
@@ -672,16 +794,20 @@ pub trait EvalContextExt<'mir, 'tcx: 'mir>: crate::MiriEvalContextExt<'mir, 'tcx
             return Ok(());
         }
         // We need this to be in-memory to use tagged pointers.
-        let return_place = this.force_allocation(return_place)?;
+        let return_place = this.force_allocation(&return_place)?;
 
         // We have to turn the place into a pointer to use the existing code.
         // (The pointer type does not matter, so we use a raw pointer.)
         let ptr_layout = this.layout_of(this.tcx.mk_mut_ptr(return_place.layout.ty))?;
-        let val = ImmTy::from_immediate(return_place.to_ref(), ptr_layout);
+        let val = ImmTy::from_immediate(return_place.to_ref(this), ptr_layout);
         // Reborrow it.
-        let val = this.retag_reference(val, RefKind::Unique { two_phase: false }, /*protector*/ true)?;
+        let val = this.retag_reference(
+            &val,
+            RefKind::Unique { two_phase: false },
+            /*protector*/ true,
+        )?;
         // And use reborrowed pointer for return place.
-        let return_place = this.ref_to_mplace(val)?;
+        let return_place = this.ref_to_mplace(&val)?;
         this.frame_mut().return_place = Some(return_place.into());
 
         Ok(())
