@@ -1,9 +1,13 @@
 fn main() {
-    #[cfg(target_pointer_width="64")]
+    #[cfg(all(target_endian="little", target_pointer_width="64"))]
     let bad = unsafe {
         std::mem::transmute::<u128, &[u8]>(42)
     };
-    #[cfg(target_pointer_width="32")]
+    #[cfg(all(target_endian="big", target_pointer_width="64"))]
+    let bad = unsafe {
+        std::mem::transmute::<u128, &[u8]>(42 << 64)
+    };
+    #[cfg(all(target_endian="little", target_pointer_width="32"))]
     let bad = unsafe {
         std::mem::transmute::<u64, &[u8]>(42)
     };
